//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/JobModel.cpp
//! @brief     Implements class JobModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/JobModel.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/PointwiseAxisItem.h"
#include "GUI/Model/Data/RealDataItem.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Job/JobModelFunctions.h"
#include "GUI/Model/Job/JobQueueData.h"
#include "GUI/Model/Job/ParameterTreeUtils.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/MultiLayerItem.h"

JobModel::JobModel(QObject* parent)
    : SessionModel(GUI::Session::XML::JobModelTag, parent), m_queue_data(nullptr)
{
    m_queue_data = new JobQueueData(this);
    connect(m_queue_data, &JobQueueData::focusRequest, this, &JobModel::focusRequest);
    connect(m_queue_data, &JobQueueData::globalProgress, this, &JobModel::globalProgress);
    setObjectName(GUI::Session::XML::JobModelTag);
}

JobModel::~JobModel()
{
    delete m_queue_data;
}

JobItem* JobModel::getJobItemForIdentifier(const QString& identifier)
{
    for (auto* jobItem : jobItems())
        if (jobItem->getIdentifier() == identifier)
            return jobItem;

    return nullptr;
}

//! Main method to add a job
JobItem* JobModel::addJob(const MultiLayerItem* multiLayerItem,
                          const InstrumentItem* instrumentItem, const RealDataItem* realDataItem,
                          const SimulationOptionsItem& optionItem)
{
    ASSERT(multiLayerItem);
    ASSERT(instrumentItem);

    auto* jobItem = insertItem<JobItem>();
    jobItem->setItemName(generateJobName());
    jobItem->setIdentifier(QUuid::createUuid().toString());

    GUI::Model::JobFunctions::setupJobItemSampleData(jobItem, multiLayerItem);
    GUI::Model::JobFunctions::setupJobItemInstrument(jobItem, instrumentItem);

    jobItem->copySimulationOptionsIntoJob(optionItem);

    GUI::Model::ParameterTreeUtils::createParameterTree(jobItem, true);

    GUI::Model::JobFunctions::setupJobItemOutput(jobItem);

    if (realDataItem)
        GUI::Model::JobFunctions::setupJobItemForFit(jobItem, realDataItem);

    return jobItem;
}

QVector<JobItem*> JobModel::jobItems() const
{
    return topItems<JobItem>();
}

//! restore instrument and sample model from backup for given JobItem
void JobModel::restore(JobItem* jobItem)
{
    jobItem->parameterContainerItem()->restoreBackupValues();
}

bool JobModel::hasUnfinishedJobs()
{
    if (m_queue_data->hasUnfinishedJobs())
        return true;

    for (auto* jobItem : jobItems())
        if (jobItem->getStatus() == JobStatus::Fitting)
            return true;

    return false;
}

void JobModel::clear()
{
    for (auto* job : jobItems())
        removeJob(job);

    SessionModel::clear();
}

QVector<SessionItem*> JobModel::nonXMLItems() const
{
    QVector<SessionItem*> result;

    for (auto* jobItem : topItems<JobItem>()) {
        if (auto* intensityItem = jobItem->dataItem())
            result.push_back(intensityItem);

        if (auto* real_data = dynamic_cast<RealDataItem*>(jobItem->realDataItem())) {
            if (auto* data_item = real_data->dataItem())
                result.push_back(data_item);
            if (auto* native_data = real_data->nativeData())
                result.push_back(native_data);
        }

        auto* instrument = dynamic_cast<SpecularInstrumentItem*>(jobItem->instrumentItem());
        if (instrument) {
            auto* axis_group = instrument->beamItem()->inclinationAxisGroup();
            result.push_back(axis_group->firstChildOfType<PointwiseAxisItem>());
        }
    }

    return result;
}

void JobModel::readFrom(QXmlStreamReader* reader, MessageService* messageService /*= 0*/)
{
    SessionModel::readFrom(reader, messageService);

    // Set & check instrument links. JobItem and realDataItem have to reference the same instrument
    // (which is the one contained in this job item, not contained in the instrument model)
    for (auto* jobItem : jobItems()) {
        if (RealDataItem* refItem = jobItem->realDataItem()) {
            refItem->setInstrumentId(jobItem->instrumentItem()->id());
            ASSERT(refItem->linkedInstrument() == jobItem->instrumentItem());
        }

        // initialize the material finder functions
        for (auto* itemWithMaterial : jobItem->sampleItem()->itemsWithMaterial())
            itemWithMaterial->setMaterialFinder([=](const QString& id) {
                return jobItem->materialItems().materialFromIdentifier(id);
            });

        // Create the not stored parameter tuning tree
        GUI::Model::ParameterTreeUtils::createParameterTree(jobItem, false);
    }
}

void JobModel::runJob(JobItem* jobItem)
{
    m_queue_data->runJob(jobItem);
}

void JobModel::cancelJob(JobItem* jobItem)
{
    m_queue_data->cancelJob(jobItem->getIdentifier());
}

void JobModel::removeJob(JobItem* jobItem)
{
    ASSERT(jobItem);
    m_queue_data->removeJob(jobItem->getIdentifier());

    removeRows(jobItem->index().row(), 1, QModelIndex());
}

QString JobModel::generateJobName()
{
    int maxJobIndex = 0;
    for (const auto* jobItem : jobItems()) {
        if (jobItem->itemName().startsWith("job"))
            maxJobIndex = std::max(maxJobIndex, jobItem->itemName().mid(3).toInt());
    }
    return QString("job%1").arg(maxJobIndex + 1);
}
