//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/MultiLayerItem.cpp
//! @brief     Implements class MultiLayerItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Session/SessionXML.h"
#include "GUI/Model/Types/VectorDescriptor.h"
#include <QXmlStreamWriter>

using namespace GUI::Session::XML;

namespace {

namespace Tags {
const QString Description("Description");
const QString Name("Name");
const QString Uid("Uid");
} // namespace Tags

} // namespace

MultiLayerItem::MultiLayerItem()
{
    m_uid = QUuid::createUuid();
    m_name = "Sample";
    m_crossCorrelationLength = 0;
}

void MultiLayerItem::initFrom(const MultiLayerItem* other)
{
    // #baMigration ++ implement
}

QVector<ItemWithMaterial*> MultiLayerItem::itemsWithMaterial() const
{
    QVector<ItemWithMaterial*> result;
    for (auto* layer : m_layers)
        result.append(layer->itemsWithMaterial());
    return result;
}

QString MultiLayerItem::sampleName() const
{
    return m_name;
}

void MultiLayerItem::setSampleName(const QString& name)
{
    m_name = name;
}

QString MultiLayerItem::description() const
{
    return m_description;
}

void MultiLayerItem::setDescription(const QString& description)
{
    m_description = description;
}

DoubleDescriptor MultiLayerItem::crossCorrLength() const
{
    DoubleDescriptor d("Cross-correlation length", &m_crossCorrelationLength, Unit::nanometer);
    d.decimals = 5;
    d.tooltip = "Cross correlation length of roughnesses between interfaces";
    return d;
}

R3 MultiLayerItem::externalField() const
{
    return m_externalField;
}

void MultiLayerItem::setExternalField(const R3& externalField)
{
    m_externalField = externalField;
}

VectorDescriptor MultiLayerItem::externalFieldVector() const
{
    VectorDescriptor d("External field", "External field (A/m)", &m_externalField, "A/m");
    d.uid = [=] { return uidForDescriptor("extField"); };
    return d;
}

QVector<LayerItem*> MultiLayerItem::layers() const
{
    return m_layers;
}

LayerItem* MultiLayerItem::addLayer(int index)
{
    if (index < 0)
        index = m_layers.size();

    auto* layer = new LayerItem();
    m_layers.insert(index, layer);
    layer->setIsTopLayer(m_layers.first() == layer);
    layer->setIsBottomLayer(m_layers.last() == layer);
    return layer;
}

void MultiLayerItem::removeLayer(LayerItem* layer)
{
    m_layers.removeAll(layer);
    delete layer;

    if (!m_layers.isEmpty()) {
        m_layers.first()->setIsTopLayer(true);
        m_layers.last()->setIsBottomLayer(true);
    }
}

void MultiLayerItem::moveLayer(LayerItem* layer, LayerItem* beforeThisLayer)
{
    m_layers.removeAll(layer);
    int index = m_layers.size(); // move to end
    if (beforeThisLayer != nullptr)
        index = m_layers.indexOf(beforeThisLayer);
    m_layers.insert(index, layer);
    for (auto l : m_layers) {
        l->setIsTopLayer(m_layers.first() == l);
        l->setIsBottomLayer(m_layers.last() == l);
    }
}

void MultiLayerItem::writeContentTo(QXmlStreamWriter* writer) const
{
    writeAttribute(writer, GUI::Session::XML::Version, int(1));
    writeAttribute(writer, Tags::Name, m_name);
    writeAttribute(writer, Tags::Description, m_description);
    writeAttribute(writer, Tags::Uid, m_uid.toString());
    writeDescriptor(writer, crossCorrLength());
    writeDescriptor(writer, externalFieldVector());

    for (auto* layer : layers()) {
        writer->writeStartElement("Layer");
        layer->writeContentTo(writer);
        writer->writeEndElement();
    }
}

void MultiLayerItem::readContentFrom(QXmlStreamReader* reader)
{
    // #baMigration ++ implement

    // set non-stored infos
    for (auto l : m_layers) {
        l->setIsTopLayer(m_layers.first() == l);
        l->setIsBottomLayer(m_layers.last() == l);
    }
}

QString MultiLayerItem::uidForDescriptor(const QString& lastPart) const
{
    return "MultiLayer/" + m_uid.toString() + "/" + lastPart;
}
